# mqtt_communication.py

# This module handles communication with the MQTT broker.

# You do not need to make any changes here.

import paho.mqtt.client as mqtt
import struct
import time
from helper import helper_get_property_names,helper_get_property_names_and_values

PublishedOutputsCounter = 0
ReceivedInputsCounter = 0
PublishOutputsInfoDone = False

# Returns the number of published output variables and received input variables.
def getMqttCounters():
    return (PublishedOutputsCounter, ReceivedInputsCounter)

# Writes output variable values to MQTT topics.
def write_outputs(client, output_variables, main_topic):
    global PublishedOutputsCounter
    nr=0
    properties=helper_get_property_names_and_values(output_variables)
    for name, value in properties.items():
        nr+=1
        # Check the type of the variable and pack the value accordingly
        if isinstance(value, bool):
            byte_message = struct.pack("?", value)
        elif isinstance(value, float):
            byte_message = struct.pack("f", value)
        elif isinstance(value, int):
            byte_message = struct.pack("h", value)
        else:
            print(
                f"Variable {name} has an unsupported type: {type(value)}"
            )
            continue
        topic = main_topic+name
        client.publish(topic, byte_message)
        if PublishedOutputsCounter==0:
          print(f"Publish output nr {nr}: {topic}")
    PublishedOutputsCounter += nr
    


# Sets up MQTT client and subscribes to input variables' topics.
def setup_mqtt(inputVariables, broker_address, port, main_topic):
    client = mqtt.Client(callback_api_version=mqtt.CallbackAPIVersion.VERSION2)

    def set_value_from_mqtt(variables_instance, topic, value):
        if main_topic != "" and main_topic in topic:
          topic = topic.split(main_topic)[1]
        if hasattr(variables_instance, topic):
            setattr(variables_instance, topic, value)
        else:
            print(f"Variable '{topic}' does not exist.")

    def on_message(client, userdata, message):
        global ReceivedInputsCounter
        topic = message.topic
        payload_length = len(message.payload)

        if payload_length == 1:
            value = struct.unpack("?", message.payload)[0]
        elif payload_length == 4:
            value = struct.unpack("f", message.payload)[0]
        elif payload_length == 2:
            value = struct.unpack("h", message.payload)[0]
        else:
            print(f"Unexpected payload length: {payload_length} bytes")
            return

        ReceivedInputsCounter += 1
        set_value_from_mqtt(inputVariables, topic, value)

    client.on_message = on_message

    client.connect(broker_address, port=port)

    input_variables = helper_get_property_names(inputVariables)
    for name in input_variables:
        # Prepend the main topic to the subscription if specified
        subscription_topic = main_topic + name
        client.subscribe(subscription_topic)
        print("Subscribed to topic: " + subscription_topic)

    client.loop_start()
    return client


if __name__ == "__main__":
    print(
        "!!!\nError: This module cannot be executed as main program. It's intended to be imported as a library.\n!!!"
    )

