# plc_prog.py

import time
from mqtt_communication import setup_mqtt, write_outputs, getMqttCounters
#from helper import helper_get_property_names

#change the following variables to match your setup
mqtt_broker_address = "localhost" #change this to the IP address of your MQTT broker if your broker is not running on the same machine
mqtt_port = 1883 #change this to the port of your MQTT broker if it is not running on the default port
main_topic="" #default: ""


#from tags import InputVariables, OutputVariables

#---------------------------------------------------------
#Insert the Input Variables and Output Variables here
#---------------------------------------------------------
#Input Variables
class InputVariables:
    S1: bool = False #Switch ON system
    S2: bool = False #Switch OFF system NC
    S3: bool = False #Start the sequence
    B1_M1Min: bool = False #M1 is at minimum
    B2_M1Max: bool = False #M1 is at maximum
    B3_M2Min: bool = False #M2 is at minimum
    B4_M2Max: bool = False #M2 is at maximum
    B5_M3Min: bool = False #M3 is at minimum
    B6_M3Max: bool = False #M3 is at maximum



#Output Variables
class OutputVariables:
    P1: bool = False #Indicator light for system is switched on.
    M1ToMin: bool = False #M1: move to minimum
    M1ToMax: bool = False #M1: move to maximum
    M2ToMin: bool = False #M2: move to minimum
    M2ToMax: bool = False #M2: move to maximum
    M3ToMin: bool = False #M3: move to minimum
    M3ToMax: bool = False #M3: move to maximum
    P2: bool = False #Lamp for sequence is active.


#---------------------------------------------------------
#Insert your PLC program here
#---------------------------------------------------------


#global variables for the PLC program (gv=global variables)
class gv:
    ControlOn = False
    Step = 0    

o: OutputVariables #type o: OutputVariables
i: InputVariables  #type i: InputVariables

# Control logic of the PLC program.
# Control logic of the PLC program.
def cyclic_control_logic(cyclic_counter: int):
    global i,o
    
    if i.S1:
        gv.ControlOn = True
    
    if not i.S2:
        gv.ControlOn = False
        gv.Step = 0

    o.P1 = gv.ControlOn
    o.P2 = (gv.Step > 0)


    # M2ToMin
    if i.S3 and gv.Step == 0:
        o.M2ToMin = True
        gv.Step += 1

    # M2 To Max
    if gv.Step == 1 and i.B3_M2Min:
        o.M2ToMin = False
        o.M2ToMax = True
        gv.Step += 1

    # M1 To Max
    if gv.Step == 2 and i.B4_M2Max:
        o.M2ToMax = False
        o.M1ToMax = True
        gv.Step += 1

    # M1 To Min and m3 to max
    if gv.Step == 3 and i.B2_M1Max:
        o.M1ToMax = False
        o.M1ToMin = True
        o.M3ToMax = True
        gv.Step += 1

    # M3 to min
    if gv.Step == 4 and i.B1_M1Min and i.B6_M3Max:
        o.M1ToMin = False
        o.M3ToMax = False
        o.M3ToMin = True
        gv.Step += 1

    # M3 to min off
    if gv.Step == 5 and i.B5_M3Min:
        o.M3ToMin = False
        gv.Step = 0

    
    
#--------------------------------------Ende cyclic_control_logic


debugTime=time.perf_counter()
# Debug log function.
# use this function to print out debug information
# to find errors in your program.
def debugLog():
    global debugTime
    if time.perf_counter() > debugTime+1:
        debugTime=time.perf_counter()
        print(f"gv.Step: {gv.Step}")
        


#---------------------------------------------------------
#---------------------------------------------------------
#---------------------------------------------------------
#Don't change the following lines

# Executes the PLC logic.
def plc_logic(cyclic_counter, inputs, outputs):
    global i,o
    i = inputs
    o = outputs
    cyclic_control_logic(cyclic_counter)
    cyclic_counter += 1

# Main function that initializes MQTT client, starts the main program loop,
# and handles exceptions.
def main():
    global mqtt_broker_address, main_topic
    # Initialisiere Input- und Output-Objekte
    instanceInputs = InputVariables()
    instanceOutputs = OutputVariables()

    try:
        client = setup_mqtt(instanceInputs, mqtt_broker_address, 1883, main_topic)

        print("Starting program, terminate with Ctrl+C")
        cyclicCount = 0
        cyclicTimeInMs = 0
        lastTime = time.perf_counter()  # Erfasse Startzeit

        while True:
            plc_logic(cyclicCount, instanceInputs, instanceOutputs)  # Führe PLC Logik aus
            cyclicCount += 1
            write_outputs(client, instanceOutputs, main_topic)  # Schreibe neue Werte der Output-Variablen
            time.sleep(10/1000)  # Einstellbares Intervall 10 = 50ms Zykluszeit

            # Ausgabe alle 300 Iterationen
            if cyclicCount % 300 == 0:
                currentTime = time.perf_counter()  # Erfasse aktuelle Zeit
                cyclicTimeInMs = (currentTime - lastTime) * 1000 / 300  # Berechne durchschnittliche Zykluszeit in ms
                publishedOutputs, receivedInputs = getMqttCounters()  
                print(f"Cyclic counter: {cyclicCount}. Average cyclic time in ms: {cyclicTimeInMs:.1f}. Published output variables: {publishedOutputs}. Received input variables: {receivedInputs}.")
                lastTime = currentTime  # Aktualisiere lastTime für die nächste Berechnung
    except KeyboardInterrupt:
        print("Program is terminated")
        client.loop_stop()
        client.disconnect()


#---------------------------------------------------------
if __name__ == "__main__":
    main()
