# plc_prog.py

import time
from mqtt_communication import setup_mqtt, write_outputs, getMqttCounters

#change the following variables to match your setup
mqtt_broker_address = "localhost" #change this to the IP address of your MQTT broker if your broker is not running on the same machine
mqtt_port = 1883 #change this to the port of your MQTT broker if it is not running on the default port
main_topic="" #default: ""


#---------------------------------------------------------
#Insert the Input Variables and Output Variables here
#---------------------------------------------------------


class InputVariables:
    B1_X_VordereEndlage: bool = False #Vordere Endlage der X-Achse erreicht
    B2_X_ReferenzPos: bool = False #Referenzposition der X-Achse erreicht
    B3_X_HintereEndlage: bool = False #Hintere Endlage der X-Achse erreicht
    B4_Y_VordereEndlage: bool = False #Vordere Endlage der Y-Achse erreicht
    B5_Y_ReferenzPos: bool = False #Referenzposition der Y-Achse erreicht
    B6_Y_HintereEndlage: bool = False #Hintere Endlage der Y-Achse erreicht
    B7_Z_ObereEndlage: bool = False #Vordere Endlage der Z-Achse erreicht
    B8_Z_ReferenzPos: bool = False #Referenzposition der Z-Achse erreicht
    B9_Z_UntereEndlage: bool = False #Hintere Endlage der Z-Achse erreicht
    B11_MagnetIsOn: bool = False #Rückmeldung Magnet ist eingeschaltet
    S1_Start: bool = False #S1 Start
    S2_Stop: bool = False #S2 Stop
    S3_ACK: bool = False #S3 Ok
    S4_REF: bool = False #S4 Referenz
    S6_1_MAN: bool = False #S6-1 Handbetrieb
    S6_2_Auto: bool = False #S6-2 Automatik
    C10_Status: int = 0 #C10 RFID Status-Byte
    C10_DataIn: int = 0 #C10 RFID Daten (Word) zyklisch gelesen
    C10_TP: bool = False #RFID Tag in der Nähe des Schreiblesekopfes
    C10_CmdEnd: bool = False #RFID Aktion oder Kommando abgeschlossen
    M1_X_HomingDone: bool = False #M1-X: HomingDone
    M1_X_HomingBusy: bool = False #M1-X: HomingBusy
    M1_X_PositioningDone: bool = False #M1-X: PositioningDone
    M1_X_PostitioningBusy: bool = False #M1-X: PositioningBusy
    M1_X_IstPosition: float = 0 #M1-X: CurrentPosition
    M3_Z_HomingDone: bool = False #M3-Z: HomingDone
    M3_Z_HomingBusy: bool = False #M3-Z: HomingBusy
    M3_Z_PositioningDone: bool = False #M3-Z: PositioningDone
    M3_Z_PostitioningBusy: bool = False #M3-Z: PositioningBusy
    M3_Z_IstPosition: float = 0 #M3-Z: CurrentPosition
    M2_Y_HomingDone: bool = False #M2-Y: HomingDone
    M2_Y_HomingBusy: bool = False #M2-Y: HomingBusy
    M2_Y_PositioningDone: bool = False #M2-Y: PositioningDone
    M2_Y_PostitioningBusy: bool = False #M2-Y: PositioningBusy
    M2_Y_IstPosition: float = 0 #M2-Y: CurrentPosition
    M1_X_StatusBitsEnable: bool = False #M1-X: StatusBitsEnable
    M1_X_StatusBitsError: bool = False #M1-X: StatusBitsError
    M1_X_StatusBitsHomingDone: bool = False #M1-X: StatusBitsHomingDone
    M2_Y_StatusBitsEnable: bool = False #M2-Y: StatusBitsEnable
    M2_Y_StatusBitsError: bool = False #M2-Y: StatusBitsError
    M2_Y_StatusBitsHomingDone: bool = False #M2-Y: StatusBitsHomingDone
    M3_Z_StatusBitsEnable: bool = False #M3-Z: StatusBitsEnable
    M3_Z_StatusBitsError: bool = False #M3-Z: StatusBitsError
    M3_Z_StatusBitsHomingDone: bool = False #M3-Z: StatusBitsHomingDone
    SimulationIstAktiv: bool = False #Simulation ist aktiv 


#Output Variables
class OutputVariables:
    M1_DIR_X: bool = False #Vorgabe der Richtung für die X-Achse
    M2_DIR_Y: bool = False #Vorgabe der Richtung für die Y-Achse
    M3_DIR_Z: bool = False #Vorgabe der Richtung für die Z-Achse
    M11_MagnetOn: bool = False #Haltemagnet einschalten
    M1_Puls_X: bool = False #Vorgabe der Takte für die X-Achse
    M2_Puls_Y: bool = False #Vorgabe der Takte für die Y-Achse
    M3_Puls_Z: bool = False #Vorgabe der Takte für die Z-Achse
    M1_EN_X: bool = False #Servo für die X-Achse einschalten
    M2_EN_Y: bool = False #Servo für die Y-Achse einschalten
    M3_EN_Z: bool = False #Servo für die Z-Achse einschalten
    P1_StartGruen: bool = False #P1 Start
    P2_StopRot: bool = False #P2 Stop
    P3_AckBlau: bool = False #P3 Ok
    P4_RefWeiss: bool = False #P4 Referenz
    C10_CmdNo: int = 0 #RFID Kommando (z.B. AUTO_READ=1
    C10_Command: int = 0 #RFID Kommando Zusatzdaten
    C10_DataOut: int = 0 #RFID Daten (Word) zyklisch geschrieben
    C10_StartCmd: bool = False #RFID Kommando jetzt ausführen
    M1_X_JogForward: bool = False #M1-X: JogForward
    M1_X_JogBackward: bool = False #M1-X: JogBackward
    M1_X_StartHoming: bool = False #M1-X: StartHoming
    M1_X_StartPositioning: bool = False #M1-X: StartPositioning
    M1_X_SollPosition: float = 0 #M1-X: SetPosition
    M1_X_Velocity: float = 0 #M1-X: Velocity
    M3_Z_JogForward: bool = False #M3-Z: JogForward
    M3_Z_JogBackward: bool = False #M3-Z: JogBackward
    M3_Z_StartHoming: bool = False #M3-Z: StartHoming
    M3_Z_StartPositioning: bool = False #M3-Z: StartPositioning
    M3_Z_SollPosition: float = 0 #M3-Z: SetPosition
    M3_Z_Velocity: float = 0 #M3-Z: Velocity
    M2_Y_JogForward: bool = False #M2-Y: JogForward
    M2_Y_JogBackward: bool = False #M2-Y: JogBackward
    M2_Y_StartHoming: bool = False #M2-Y: StartHoming
    M2_Y_StartPositioning: bool = False #M2-Y: StartPositioning
    M2_Y_SollPosition: float = 0 #M2-Y: SetPosition
    M2_Y_Velocity: float = 0 #M2-Y: Velocity
    ShowSortingSlideFromPlc: bool = False #
    RandomAutoCreate: bool = False #


#---------------------------------------------------------
#Insert your PLC program here
#---------------------------------------------------------

import threading

class MotionAxis:
    def __init__(self, inputVariables, outputVariables, startOutputPropertyName, targetPositionOutputPropertyName, doneOutputPropertyName):
        self.InstanceInputVariables = inputVariables
        self.InstanceOutVariables = outputVariables
        self.StartOutputPropertyName = startOutputPropertyName
        self.TargetPositionOutputPropertyName = targetPositionOutputPropertyName
        self.DoneOutputPropertyName = doneOutputPropertyName
        self.moveToStartTime=time.perf_counter()        

    # Setzt das Ziel der Bewegung
    def SetTarget(self, value):
        setattr(self.InstanceOutVariables, self.TargetPositionOutputPropertyName, value)

    # Startet oder stoppt die Bewegung
    def SetValueStart(self, value):
        setattr(self.InstanceOutVariables, self.StartOutputPropertyName, value)

    def GetValueStart(self):
        getattr(self.InstanceOutVariables, self.StartOutputPropertyName)

    # Gibt zurück, ob die Bewegung abgeschlossen ist
    def GetValueDone(self):
        return getattr(self.InstanceInputVariables, self.DoneOutputPropertyName)

    # Startet die Bewegung zur Zielposition
    def MoveTo(self, targetPosition):
        #aktuelle Zeit merken
        self.moveToStartTime=time.perf_counter()                
        self.SetTarget(targetPosition)
        self.SetValueStart(False)                
        # Verzögerung 300ms, beim Starten, damit eine Hi-Flanke am Antrieb ankommt
        threading.Timer(0.3, self.SetValueStart, [True]).start()
        

    # Prüft, ob die Bewegung abgeschlossen ist
    def IsReady(self):
        if time.perf_counter()<self.moveToStartTime+0.3: #nicht sofort nach dem Starten prüfen!
            return False        
        
        if self.GetValueDone():            
            return True       
        
        return False

class PickAndPlaceHelper:
    def __init__(self, x_axis, y_axis, z_axis, outputVariables, magnetOutputPropertyName, topPositionZ):
        self.topPositionZ=topPositionZ
        self.outputVariables = outputVariables
        self.x_axis = x_axis
        self.y_axis = y_axis
        self.z_axis = z_axis
        self.magnetOutputPropertyName = magnetOutputPropertyName        
        self.pickPositionZ=0
        self.placePositionX=0
        self.placePositionY=0
        self.placePositionZ=0        
        self.status="ready"

    # Magnet ein- und ausschalten
    def Magnet(self, value):
        setattr(self.outputVariables, self.magnetOutputPropertyName, value)


    # Startet die Bewegung zur Zielposition
    def Start(self, pickPositionX, pickPositionY, pickPositionZ, placePositionX, placePositionY, placePositionZ):
        #aktuelle Zeit merken
        self.status="pick0" #nach x,y fahren (z bleibt gleich)
        print(f"Start ---------------------------- Pick0 Move to {pickPositionX},{pickPositionY},{self.topPositionZ}")
        self.x_axis.MoveTo(pickPositionX)
        self.y_axis.MoveTo(pickPositionY)
        self.z_axis.MoveTo(self.topPositionZ)#obere Pos, bei Fahrweg        
        self.pickPositionZ=pickPositionZ
        self.placePositionX=placePositionX
        self.placePositionY=placePositionY
        self.placePositionZ=placePositionZ
        

    # Prüft, ob die Bewegung abgeschlossen ist
    def IsReady(self):        
        
        #ist an x-y-position angekommen? Dann nach unten fahren
        if self.status=="pick0":
           if self.x_axis.IsReady() and self.y_axis.IsReady() and self.z_axis.IsReady():
              print(f"Status={self.status} Move to {self.pickPositionZ}")
              self.z_axis.MoveTo(self.pickPositionZ)              
              self.status="pick1"              
           return False
                
        #unten angekommen, dann wieder hochfahren
        elif self.status=="pick1":#z nach oben fahren
           if self.z_axis.IsReady():              
              print(f"Status={self.status} Move to {self.topPositionZ}")
              self.Magnet(True)
              self.z_axis.MoveTo(self.topPositionZ)
              self.status="pick2"
           return False

        #oben angekommen, dann zur place x-y pos fahren
        elif self.status=="pick2":
           if self.z_axis.IsReady():
              print(f"Status={self.status} Move to {self.placePositionX},{self.placePositionY}")
              self.x_axis.MoveTo(self.placePositionX)
              self.y_axis.MoveTo(self.placePositionY)              
              self.status="place0"      
           return False

        #place position angekommen, x-y? dann nach unten fahren
        elif self.status=="place0":
           if self.x_axis.IsReady() and self.y_axis.IsReady():
              print(f"Status={self.status} Move to {self.placePositionZ}")
              self.z_axis.MoveTo(self.placePositionZ)
              self.status="place1"
           return False

        #unten angekommen, dann wieder hochfahren
        elif self.status=="place1":
           if self.z_axis.IsReady():
              print(f"Status={self.status} Move to {self.topPositionZ}")
              self.Magnet(False)
              self.z_axis.MoveTo(self.topPositionZ)
              self.status="place2"
           return False

        #oben angekommen? dann fertig
        elif self.status=="place2":
           if self.z_axis.IsReady():
              print(f"Status={self.status} - fertig")
              self.status="ready"
              #self.isBusy=False
        
        return self.status=="ready"



#global variables for the PLC program (gv=global variables)
class gv:
    ControlOn = False
    Step = 0    
    x_axis = None  # Placeholder für MotionAxis
    y_axis = None  # Placeholder für MotionAxis
    z_axis = None  # Placeholder für MotionAxis
    pickAndPlace = None  # Placeholder für PickAndPlaceHelper
    count_white = 0
    count_black = 0
    count_green = 0
    pickXPosRfId=-7
    pickYPosRfId=265
    placeXPosStart = 190
    placeYPosStart = 110
    placeXPosDelete=110
    placeYPosDelete=265
    topZMovePosition = -10
    itemWith = 50
    itemHeight = 23
    maxRows = 3
    PickDownPos = 45


o: OutputVariables #type o: OutputVariables
i: InputVariables  #type i: InputVariables


def initMyPlcPrg():
    global i,o
    gv.z_axis = MotionAxis(i, o, "M3_Z_StartPositioning", "M3_Z_SollPosition", "M3_Z_PositioningDone")
    gv.x_axis = MotionAxis(i, o, "M1_X_StartPositioning", "M1_X_SollPosition", "M1_X_PositioningDone")
    gv.y_axis = MotionAxis(i, o, "M2_Y_StartPositioning", "M2_Y_SollPosition", "M2_Y_PositioningDone")
    gv.pickAndPlace = PickAndPlaceHelper(gv.x_axis, gv.y_axis, gv.z_axis, o, "M11_MagnetOn", gv.topZMovePosition)
    o.M1_EN_X = True #Achse einschalten
    o.M2_EN_Y = True #Achse einschalten
    o.M3_EN_Z = True #Achse einschalten
    o.C10_CmdNo = 1  # AutoRead - immer lesen
    o.C10_StartCmd = True #Starte RFID Kommando


#berechnet die x und z position für ein item aufgrund der Anzahl der items
def getXZPlacePosition(itemCount: int):
    xPosition = gv.placeXPosStart+itemCount % gv.maxRows * gv.itemWith
    layer = itemCount // gv.maxRows #"//" ist ganzzahlige Division
    zPosition = gv.PickDownPos - layer * gv.itemHeight 
    return xPosition, zPosition    #Rückgabe als Tupel

# Control logic of the PLC program.
def cyclic_control_logic(cyclic_counter: int):
    global i,o
        
    debugLog()
    
    if cyclic_counter==0:
        initMyPlcPrg() #einmalige Initialisierung, wenn cyclic_counter=0

    o.RandomAutoCreate=not o.RandomAutoCreate; #toggle RandomAutoCreate, dadurch werden immer automatisch neue items erstellt
    
    #yPos der verschiedenen items
    _placeYPosGreen=gv.placeYPosStart+gv.itemWith*2 #y-pos der grünen items
    _placeYPosBlack=gv.placeYPosStart+gv.itemWith #y-pos der schwarzen items
    _placeYPosWhite=gv.placeYPosStart #y-pos der weißen items

    _placeXPos=_placeDownPos=_placeYPos=0


    #start wenn der RFID Sensor ein item erkennt
    if gv.Step == 0 and i.C10_TP:
        print("pickAndPlace.Start")       
        if i.C10_DataIn==1: #weiß
            _placeYPos=_placeYPosWhite            
            (_placeXPos,_placeDownPos)=getXZPlacePosition(gv.count_white)            
            gv.count_white+=1
        if i.C10_DataIn==2: #schwarz
            #_placeYPos=_placeYPosBlack
            #(_placeXPos,_placeDownPos)=getXZPlacePosition(gv.count_black)                        
            gv.count_black+=1
            #schwarze items werden nicht benötigt und werden gelöscht ->an die Löschposition transportieren
            _placeXPos=gv.placeXPosDelete
            _placeYPos=gv.placeYPosDelete
            _placeDownPos=gv.topZMovePosition
        if i.C10_DataIn==3: #grün
            _placeYPos=_placeYPosGreen
            (_placeXPos,_placeDownPos)=getXZPlacePosition(gv.count_green)            
            gv.count_green+=1

        #Begrenzung der z-position
        if _placeDownPos<gv.topZMovePosition:
            _placeDownPos=gv.topZMovePosition
        
        #starte den pick and place vorgang
        gv.pickAndPlace.Start(gv.pickXPosRfId,gv.pickYPosRfId,gv.PickDownPos,  _placeXPos,_placeYPos,_placeDownPos)
        gv.Step+=1 #nächster schritt
    
    elif gv.Step==1 and gv.pickAndPlace.IsReady(): #wenn der pick and place vorgang abgeschlossen ist
        gv.Step=0  
    
#--------------------------------------Ende cyclic_control_logic
#-------------------------------------------------------------------------------------------------
#-------------------------------------------------------------------------------------------------

debugTime=time.perf_counter()
# Debug log function.
# use this function to print out debug information
# to find errors in your program.
def debugLog():
    global debugTime
    if time.perf_counter() > debugTime+1:
        debugTime=time.perf_counter()
        print(f"Step: {gv.Step}")
        


#---------------------------------------------------------
#---------------------------------------------------------
#---------------------------------------------------------
#Don't change the following lines

# Executes the PLC logic.
def plc_logic(cyclic_counter, inputs, outputs):
    global i,o
    i = inputs
    o = outputs
    cyclic_control_logic(cyclic_counter)
    cyclic_counter += 1

# Main function that initializes MQTT client, starts the main program loop,
# and handles exceptions.
def main():
    global mqtt_broker_address, main_topic
    # Initialisiere Input- und Output-Objekte
    instanceInputs = InputVariables()
    instanceOutputs = OutputVariables()

    try:
        client = setup_mqtt(instanceInputs, mqtt_broker_address, 1883, main_topic)

        print("Starting program, terminate with Ctrl+C")
        cyclicCount = 0
        cyclicTimeInMs = 0
        lastTime = time.perf_counter()  # Erfasse Startzeit

        while True:
            plc_logic(cyclicCount, instanceInputs, instanceOutputs)  # Führe PLC Logik aus
            cyclicCount += 1
            write_outputs(client, instanceOutputs, main_topic)  # Schreibe neue Werte der Output-Variablen
            time.sleep(50/1000)  # Einstellbares Intervall 50 = 50ms Zykluszeit

            # Ausgabe alle 300 Iterationen
            if cyclicCount % 300 == 0:
                currentTime = time.perf_counter()  # Erfasse aktuelle Zeit
                cyclicTimeInMs = (currentTime - lastTime) * 1000 / 300  # Berechne durchschnittliche Zykluszeit in ms
                publishedOutputs, receivedInputs = getMqttCounters()  
                print(f"Cyclic counter: {cyclicCount}. Average cyclic time in ms: {cyclicTimeInMs:.1f}. Published output variables: {publishedOutputs}. Received input variables: {receivedInputs}.")
                lastTime = currentTime  # Aktualisiere lastTime für die nächste Berechnung
    except KeyboardInterrupt:
        print("Program is terminated")
        client.loop_stop()
        client.disconnect()


#---------------------------------------------------------
if __name__ == "__main__":
    main()
